//
// Copyright (c) 2002
// Ronald Kevin Burton
//
// This material is provided "as is", with absolutely no warranty expressed
// or implied. This software is provided solely as part of ".NET Common Language
// Runtime Unleashed" to help illustrate concepts and principles described in 
// that book. Any use is at your own risk.
//
// Permission to use or copy this software for any purpose is hereby granted 
// without fee, provided the above notices are retained on all copies.
// Permission to modify the code and to distribute modified code is granted,
// provided the above notices are retained, and a notice that the code was
// modified is included with the above copyright notice.
//
// 
// SEHException.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include <string>
#include <iostream>
#include <stdexcept>

int EvalException ( int exception, LPEXCEPTION_POINTERS pinfo)
{
	// Pass on most exceptions
	switch(exception)
	{
	case EXCEPTION_ACCESS_VIOLATION:
		std::cout << "EXCEPTION_ACCESS_VIOLATION 0x" << std::hex << exception << std::endl;
		break;
	case EXCEPTION_BREAKPOINT:
		std::cout << "EXCEPTION_BREAKPOINT 0x" << std::hex  << exception << std::endl;
		break;
	case EXCEPTION_DATATYPE_MISALIGNMENT:
		std::cout << "EXCEPTION_DATATYPE_MISALIGNMENT 0x" << std::hex  << exception << std::endl;
		break;
	case EXCEPTION_SINGLE_STEP:
		std::cout << "EXCEPTION_SINGLE_STEP 0x" << std::hex  << exception << std::endl;
		break;
	case EXCEPTION_ARRAY_BOUNDS_EXCEEDED:
		std::cout << "EXCEPTION_ARRAY_BOUNDS_EXCEEDED 0x" << std::hex  << exception << std::endl;
		break;
	case EXCEPTION_FLT_DENORMAL_OPERAND:
		std::cout << "EXCEPTION_FLT_DENORMAL_OPERAND 0x" << std::hex  << exception << std::endl;
		break;
	case EXCEPTION_FLT_DIVIDE_BY_ZERO:
		std::cout << "EXCEPTION_FLT_DIVIDE_BY_ZERO 0x" << std::hex  << exception << std::endl;
		break;
	case EXCEPTION_FLT_INEXACT_RESULT:
		std::cout << "EXCEPTION_FLT_INEXACT_RESULT 0x" << std::hex  << exception << std::endl;
		break;
	case EXCEPTION_FLT_INVALID_OPERATION:
		std::cout << "EXCEPTION_FLT_INVALID_OPERATION 0x" << std::hex  << exception << std::endl;
		break;
	case EXCEPTION_FLT_OVERFLOW:
		std::cout << "EXCEPTION_FLT_OVERFLOW 0x" << std::hex  << exception << std::endl;
		break;
	case EXCEPTION_FLT_STACK_CHECK:
		std::cout << "EXCEPTION_FLT_STACK_CHECK 0x" << std::hex  << exception << std::endl;
		break;
	case EXCEPTION_FLT_UNDERFLOW:
		std::cout << "EXCEPTION_FLT_UNDERFLOW 0x" << std::hex << exception << std::endl;
		break;
	case EXCEPTION_INT_DIVIDE_BY_ZERO:
		std::cout << "EXCEPTION_INT_DIVIDE_BY_ZERO 0x" << std::hex << exception << std::endl;
		break;
	case EXCEPTION_PRIV_INSTRUCTION:
		std::cout << "EXCEPTION_PRIV_INSTRUCTION 0x" << std::hex  << exception << std::endl;
		break;
	case EXCEPTION_NONCONTINUABLE_EXCEPTION:
		std::cout << "EXCEPTION_NONCONTINUABLE_EXCEPTION 0x" << std::hex  << exception << std::endl;
		break;
	}

	std::cout << "Code: 0x" << std::hex << pinfo->ExceptionRecord->ExceptionCode << std::endl;
	std::cout << "Flags: " << ((pinfo->ExceptionRecord->ExceptionFlags & EXCEPTION_NONCONTINUABLE) ? "Not-Continuable" : "Continuable")  << std::endl;
	std::cout << "Exception Record: 0x" << std::hex << pinfo->ExceptionRecord->ExceptionRecord  << std::endl;
	std::cout << "Address: 0x" << std::hex << pinfo->ExceptionRecord->ExceptionAddress << std::endl;
	std::cout << "Number of Parameters: " << pinfo->ExceptionRecord->NumberParameters << std::endl;
	if(exception == EXCEPTION_ACCESS_VIOLATION)
	{
		if(pinfo->ExceptionRecord->ExceptionInformation[0] == 0)
			std::cout << "Attempt to read inaccessible data 0x" << std::hex << pinfo->ExceptionRecord->ExceptionInformation[1] << std::endl;
		else if(pinfo->ExceptionRecord->ExceptionInformation[0] == 1)
			std::cout << "Attempt to write inaccessible data at 0x" << std::hex << pinfo->ExceptionRecord->ExceptionInformation[1] << std::endl;
	}

	// EXCEPTION_EXECUTE_HANDLER       1
	// EXCEPTION_CONTINUE_SEARCH       0
	// EXCEPTION_CONTINUE_EXECUTION    -1
    return EXCEPTION_EXECUTE_HANDLER;
}
class SpecialClass
{
public:
	void Output()
	{
		std::cout << "Calling SpecialClass::Output()" << std::endl;
	}
};

void SEHExceptions()
{
	__try
	{
		int a = 0;
		int b = 1;
		int c = b/a;
		std::cout << "The result is: " << c << std::endl;
	}
	__except( EvalException(GetExceptionCode(), GetExceptionInformation()) )
	{
		std::cout << "Caught exception" << std::endl;
	}

	__try
	{
		int *p = 0;
		std::cout << "The value is: " << p[1] << std::endl;
	}
	__except( EvalException(GetExceptionCode(), GetExceptionInformation()) )
	{
		std::cout << "Caught exception" << std::endl;
	}

	__try
	{
		SpecialClass *p = 0;
		p->Output();
	}
	__except( EvalException(GetExceptionCode(), GetExceptionInformation()) )
	{
		std::cout << "Caught exception" << std::endl;
	}
}

class custom_exception : public std::exception
{
public:
	explicit custom_exception(const std::string& _message) : message(_message)
	{
	}

	virtual ~custom_exception()
	{
	}

	virtual const char *what()
	{
		return (message.c_str());
	}

private:
	// The message
	std::string message;
};


void EHExceptions()
{
	try
	{
		int a = 0;
		int b = 1;
		int c = b/a;
		std::cout << "The result is: " << c << std::endl;
	}
	catch(std::exception)
	{
		std::cout << "Caught EH exception " << std::endl;
	}
	catch(...)
	{
		std::cout << "Default caught EH exception " << std::endl;
	}

	try
	{
		int *p = 0;
		std::cout << "The value is: " << p[1] << std::endl;
	}
	catch(std::exception)
	{
		std::cout << "Caught EH exception " << std::endl;
	}
	catch(...)
	{
		std::cout << "Default caught EH exception " << std::endl;
	}
	try
	{
		throw custom_exception("This is a test");
	}
	catch(custom_exception& e)
	{
		std::cout << "Caught custom exception: " <<  e.what() << std::endl;
	}
}


int _tmain(int argc, _TCHAR* argv[])
{
	SEHExceptions();
	EHExceptions();
	return 0;
}

